/*
 * Copyright (C) 2024 sillysagiri
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
 */
 
#pragma once

#include "stb_image.h"
#include <cstdint>
#include <filesystem>
#include <string>
#include <vector>

// wporkaround surpress clangd pragma pack warning
// https://github.com/clangd/clangd/issues/1167
static_assert(true);

#pragma pack(push, 1)
struct Metadata {
	uint64_t header = 0x676D69796C6C6973;
	uint8_t version = 13;
	uint8_t format;
	uint16_t width;
	uint16_t height;
	uint16_t palette_count;
	uint32_t palette_hash;
	uint32_t original_size;
	uint32_t compress_size;
};
#pragma pack(pop)

enum Format
{
	Format_RGB_256 = 0,
	Format_RGB_16,
	Format_INDEXED_4,
	Format_INDEXED_16,
	Format_INDEXED_256,
	Format_INDEXED_32A8,
	Format_INDEXED_8A32,
	Format_PALETTE_16,
};

struct Color {
	double r = 0;
	double g = 0;
	double b = 0;
	double a = 255;

	inline uint16_t toRGB16(uint8_t alpha_threshold = 128) const {
		const uint16_t alpha_bit = (a >= alpha_threshold) ? 0b1000000000000000 : 0;
		return alpha_bit
				| (((uint8_t)r >> 3) & 0b00011111)
				| (((uint8_t)g >> 3) & 0b00011111) << 5
				| (((uint8_t)b >> 3) & 0b00011111) << 10;
	}
};

struct Image {
	int width = 0, height = 0;
	std::vector<Color> data;

	~Image() {}
	void Load(const std::filesystem::path &path);
};

struct Options {
	std::vector<std::string> path_input;
	std::string path_output;
	std::string format;
};

void Verify();
void Convert();

std::vector<Color> QuantizePalette(const int numColor, const Image &image);
std::vector<uint8_t> RemapImage(const Image &image, const std::vector<Color> &palette, const int &format);

std::vector<Color> inc_online_kmeans ( const Image *img, const int num_colors, 
		    const double lr_exp = 0.5, const double sample_rate = 0.5);

/* Max. L_2^2 distance in 24-bit RGB space = 3 * 255 * 255 */
#define MAX_RGB_DIST 195075